from barbell2.bodycomp.totalseg import TotalSegmentator
from barbell2.bodycomp.selectroi import RoiSelector
from barbell2.bodycomp.selectslice import SliceSelector
from barbell2.bodycomp.seg import MuscleFatSegmentator
from barbell2.bodycomp.calculator import BodyCompositionCalculator



import os
import json
import zipfile
import logging
import pydicom
import numpy as np

from barbell2.utils import is_dicom_file, get_pixels

logger = logging.getLogger(__name__)


class MuscleFatSegmentator:

    ARGMAX = 0
    PROBABILITIES = 1

    def __init__(self):
        self.input_files = None
        self.image_dimensions = None
        self.model_files = None
        self.mode = MuscleFatSegmentator.ARGMAX
        self.output_directory = None
        self.output_segmentation_files = None

    @staticmethod
    def load_model(file_path):
        import tensorflow as tf
        file_name = os.path.split(file_path)[1]
        model_directory = '/tmp/barbell2/bodycomp/seg.py'
        os.makedirs(model_directory, exist_ok=True)
        with zipfile.ZipFile(file_path) as zip_obj:
            zip_obj.extractall(path=model_directory)
        return tf.keras.models.load_model(model_directory, compile=False)

    @staticmethod
    def load_params(file_path):
        with open(file_path, 'r') as f:
            return json.load(f)

    def load_model_files(self):
        model, contour_model, params = None, None, None
        for file_path in self.model_files:
            file_name = os.path.split(file_path)[1]
            if file_name == 'model.zip':
                model = self.load_model(file_path)
            elif file_name == 'contour_model.zip':
                contour_model = self.load_model(file_path)
            elif file_name == 'params.json':
                params = self.load_params(file_path)
            else:
                logger.error(f'Unknown model file {file_name}')
        return model, contour_model, params

    @staticmethod
    def normalize(img, min_bound, max_bound):
        img = (img - min_bound) / (max_bound - min_bound)
        img[img > 1] = 0
        img[img < 0] = 0
        c = (img - np.min(img))
        d = (np.max(img) - np.min(img))
        img = np.divide(c, d, np.zeros_like(c), where=d != 0)
        return img

    def predict_contour(self, contour_model, src_img, params):
        ct = np.copy(src_img)
        ct = self.normalize(ct, params['min_bound_contour'], params['max_bound_contour'])
        img2 = np.expand_dims(ct, 0)
        img2 = np.expand_dims(img2, -1)
        pred = contour_model.predict([img2])
        pred_squeeze = np.squeeze(pred)
        pred_max = pred_squeeze.argmax(axis=-1)
        mask = np.uint8(pred_max)
        return mask

    @staticmethod
    def convert_labels_to_157(prediction):
        new_prediction = np.copy(prediction)
        new_prediction[new_prediction == 1] = 1
        new_prediction[new_prediction == 2] = 5
        new_prediction[new_prediction == 3] = 7
        return new_prediction

    def execute(self):
        if self.input_files is None:
            logger.error('Input files not specified')
            return None
        if self.image_dimensions is None:
            logger.error('Image dimensions not specified')
            return None
        if self.model_files is None:
            logger.error('Model files not specified')
            return None
        if self.output_directory is None:
            logger.error('Output directory not specified')
            return None
        os.makedirs(self.output_directory, exist_ok=True)
        model, contour_model, params = self.load_model_files()
        self.output_segmentation_files = []
        for f in self.input_files:
            f_name = os.path.split(f)[1]
            if is_dicom_file(f):
                p = pydicom.dcmread(f)
                img1 = get_pixels(p, normalize=True)
                if contour_model is not None:
                    mask = self.predict_contour(contour_model, img1, params)
                    img1 = self.normalize(img1, params['min_bound'], params['max_bound'])
                    img1 = img1 * mask
                else:
                    img1 = self.normalize(img1, params['min_bound'], params['max_bound'])
                img1 = img1.astype(np.float32)
                img2 = np.expand_dims(img1, 0)
                img2 = np.expand_dims(img2, -1)
                pred = model.predict([img2])
                pred_squeeze = np.squeeze(pred)
                print(pred_squeeze[256][256])
                if self.mode == MuscleFatSegmentator.ARGMAX:
                    pred_max = pred_squeeze.argmax(axis=-1)
                    pred_max = self.convert_labels_to_157(pred_max)
                    segmentation_file = os.path.join(self.output_directory, f'{f_name}.seg.npy')
                    self.output_segmentation_files.append(segmentation_file)
                    np.save(segmentation_file, pred_max)
                elif self.mode == MuscleFatSegmentator.PROBABILITIES:
                    segmentation_file = os.path.join(self.output_directory, f'{f_name}.seg.prob.npy')
                    self.output_segmentation_files.append(segmentation_file)
                    np.save(segmentation_file, pred_squeeze)
                else:
                    logger.warn(f'Unknown mode {self.mode}')
            else:
                logger.warning(f'File {f} is not a valid DICOM file')
        return self.output_segmentation_files


if __name__ == '__main__':
    def main():
        segmentator = MuscleFatSegmentator()
        # segmentator.input_files = ['/Users/ralph/Desktop/SliceSelector/L3.dcm']
        segmentator.input_files = ['/mnt/localscratch/cds/rbrecheisen/raw/pancreas-demo-1/1.dcm']
        segmentator.image_dimensions = (512, 512)
        segmentator.model_files = [
            '/mnt/localscratch/cds/rbrecheisen/models/v2/model.zip',
            '/mnt/localscratch/cds/rbrecheisen/models/v2/contour_model.zip',
            '/mnt/localscratch/cds/rbrecheisen/models/v2/params.json',
        ]
        # segmentator.mode = MuscleFatSegmentator.ARGMAX
        segmentator.mode = MuscleFatSegmentator.PROBABILITIES
        segmentator.output_directory = '/tmp/barbell2/bodycomp/seg.py'
        files = segmentator.execute()
        for f in files:
            print(f)
    main()



import math
import os
import json
import datetime
import shutil

import numpy as np
import pandas as pd


class CastorVariableCoverageTableBuilder:

    def __init__(self, cast_data):
        self.cast_data = cast_data
        self.records = None

    @staticmethod
    def calculate_proportion(values):
        return np.sum(values), len(values)

    @staticmethod
    def calculate_percentage(proportion):
        percentage = proportion[0] / float(proportion[1])
        return int(percentage * 100)

    def execute(self):
        print('Building records...')
        flags = {}
        for column in self.cast_data.df.columns:
            flags[column] = []
            for idx, row in self.cast_data.df.iterrows():
                if row[column] == '' or pd.isna(row[column]):
                    flags[column].append(0)
                else:
                    flags[column].append(1)
        print('Calculating coverage...')
        data = {
            'column': [],
            'proportion': [],
            'percentage': [],
        }
        for column in flags.keys():
            data['column'].append(column)
            proportion = self.calculate_proportion(flags[column])
            data['proportion'].append('{} / {}'.format(proportion[0], proportion[1]))
            data['percentage'].append(self.calculate_percentage(proportion))
        print(json.dumps(data, indent=4))
        self.records = pd.DataFrame(data=data)
        return self.records

    def save(self, file_path):
        self.records.to_csv(file_path, index=False, sep=';')


class CastorRecordBuilder:

    def __init__(self, cast_data, dica_data):
        self.cast_data = cast_data
        self.dica_data = dica_data
        self.records = None

    def execute(self):
        print('Building records...')
        last_surgery_date = self.cast_data.get_last_surgery_date()
        indexes = []
        for idx, row in self.dica_data.df.iterrows():
            surgery_date = row['datok']
            if surgery_date > last_surgery_date:
                indexes.append(idx)
        self.records = self.dica_data.df.loc[indexes]
        # drop columns not occurring in Castor
        columns_to_drop = []
        for c in self.dica_data.df.columns:
            c_cast = self.dica_data.prefix + c
            if c_cast not in self.cast_data.df.columns:
                columns_to_drop.append(c)
        self.records = self.records.drop(columns_to_drop, axis='columns')
        nr_records = len(self.records.index)
        # update date format to dd-mm-yyyy
        for date_column in self.dica_data.get_date_columns():
            if date_column in self.records.columns:
                self.records[date_column] = self.records[date_column].dt.strftime('%d-%m-%Y')
        # rename columns with prefix dpca_
        self.records = self.records.add_prefix(self.dica_data.prefix)
        # add extra columns
        self.records.insert(0, f'{self.dica_data.prefix}idcode', self.records[f'{self.dica_data.prefix}verrichting_upn'])
        self.records.insert(0, 'Participant Creation Date', datetime.datetime.now().strftime('%d-%m-%Y'))
        self.records.insert(0, 'Site Abbreviation', 'MUMC')
        self.records.insert(0, 'Participant Status', 'Not Set')
        # add list of new participant IDs
        participant_ids = []
        next_record_nr = int(self.cast_data.get_last_record_id()[1:]) + 1
        for i in range(nr_records):
            participant_ids.append('T{:04d}'.format(next_record_nr))
            next_record_nr += 1
        self.records.insert(0, 'Participant Id', participant_ids)
        print(f'Added participant IDs: {participant_ids}')

    def save(self, output_dir):
        if os.path.isdir(output_dir):
            shutil.rmtree(output_dir)
        os.makedirs(output_dir)
        max_nr_cells = 25000
        nr_columns = len(self.records.columns)
        nr_rows = len(self.records.index)
        print(f'Saving {nr_rows} records...')
        nr_cells = nr_rows * nr_columns
        if nr_cells <= max_nr_cells:
            f_path = os.path.join(output_dir, 'new_records.csv')
            self.records.to_csv(f_path, index=False, sep=';')
            print(f'Saved {f_path}')
        else:
            nr_rows_block = int(math.floor(max_nr_cells / nr_columns))
            nr_blocks = int(math.floor(nr_rows / nr_rows_block))
            i1 = 0
            i2 = nr_rows_block
            for i in range(nr_blocks):
                df = self.records.iloc[i1:i2]
                f_path = os.path.join(output_dir, 'new_records_{:02d}.csv'.format(i))
                df.to_csv(f_path, index=False, sep=';')
                print(f'Saved {f_path}')
                i1 = i2
                i2 += nr_rows_block
            df = self.records.iloc[i1:]
            f_path = os.path.join(output_dir, 'new_records_{:02d}.csv'.format(nr_blocks))
            df.to_csv(f_path, index=False, sep=';')
            print(f'Saved {f_path}')



import os
import logging
import pydicom
import nibabel
import numpy as np

logger = logging.getLogger(__name__)


class SliceSelector:

    ALL = 0
    MEDIAN = 1
    IQR_25_50_75 = 2

    def __init__(self):
        self.input_roi = None
        self.input_volume = None
        self.input_dicom_directory = None
        self.mode = None
        self.output_files = None

    @staticmethod
    def has_duplicate_objects(roi):
        return False

    @staticmethod
    def get_min_max_slice_idx(roi):
        roi_data = roi.get_fdata()
        nr_slices = roi_data.shape[2]
        i_min = -1
        for i in range(nr_slices):
            slice = roi_data[:,:,i]
            if 1 in np.unique(slice):
                i_min = i
                break
        i_max = -1
        for i in range(nr_slices):
            slice = roi_data[:,:,nr_slices-i-1]
            if 1 in np.unique(slice):
                i_max = nr_slices - i
                break
        return i_min, i_max

    @staticmethod
    def get_z_coord_patient_position(i, volume):
        M = volume.affine[:3, :3]
        abc = volume.affine[:3, 3]
        return (M.dot([0, 0, i]) + abc)[2]

    def get_min_max_z_coord_patient_position(self, i_min, i_max, volume):
        z_min = self.get_z_coord_patient_position(i_min, volume)
        z_max = self.get_z_coord_patient_position(i_max, volume)
        return z_min, z_max

    @staticmethod
    def get_dicom_z(file_path):
        p = pydicom.dcmread(file_path, stop_before_pixels=True)
        return p.ImagePositionPatient[2]

    def get_dicom_images_between(self, z_min, z_max, dicom_directory):
        z_coords = {}
        for f in os.listdir(dicom_directory):
            f_path = os.path.join(dicom_directory, f)
            z = self.get_dicom_z(f_path)
            z_coords[z] = f_path
        z_min_nearest = min(list(z_coords.keys()), key=lambda x:abs(x - z_min))
        z_max_nearest = min(list(z_coords.keys()), key=lambda x:abs(x - z_max))
        file_paths = []
        for z in z_coords.keys():
            if z_min_nearest <= z <= z_max_nearest:
                file_paths.append(z_coords[z])
        return file_paths

    def execute(self):
        if self.input_roi is None:
            logger.error('Input ROI not specified')
            return None
        if self.input_volume is None:
            logger.error('Input NIFTI volume not specified')
            return None
        if self.input_dicom_directory is None:
            logger.error('Input DICOM directory not specified')
            return None
        if self.mode is None:
            logger.error('Mode not specified')
            return None
        roi = nibabel.load(self.input_roi)
        if self.has_duplicate_objects(roi):
            logger.error('ROI has duplicate objects')
            return None
        i_min, i_max = self.get_min_max_slice_idx(roi)
        volume = nibabel.load(self.input_volume)
        z_min, z_max = self.get_min_max_z_coord_patient_position(i_min, i_max, volume)
        self.output_files = []
        if self.mode == SliceSelector.ALL:
            self.output_files = self.get_dicom_images_between(z_min, z_max, self.input_dicom_directory)
        elif self.mode == SliceSelector.MEDIAN:
            z_median = z_min + np.abs(z_max - z_min) * 0.50
            self.output_files = self.get_dicom_images_between(z_median, z_median, self.input_dicom_directory)
        elif self.mode == SliceSelector.IQR_25_50_75:
            z_25 = z_min + np.abs(z_max - z_min) * 0.25
            self.output_files.extend(self.get_dicom_images_between(z_25, z_25, self.input_dicom_directory))
            z_50 = z_min + np.abs(z_max - z_min) * 0.50
            self.output_files.extend(self.get_dicom_images_between(z_50, z_50, self.input_dicom_directory))
            z_75 = z_min + np.abs(z_max - z_min) * 0.75
            self.output_files.extend(self.get_dicom_images_between(z_75, z_75, self.input_dicom_directory))
        elif isinstance(self.mode, float):
            z = z_min + self.mode * (z_max - z_min)
            self.output_files = self.get_dicom_images_between(z, z, self.input_dicom_directory)
        else:
            logger.error(f'Unknown mode: {self.mode}')
        return self.output_files

    def get_dicom_image_at_instance_number(self, instance_number):
        for f in os.listdir(self.input_dicom_directory):
            f_path = os.path.join(self.input_dicom_directory, f)
            p = pydicom.dcmread(f_path, stop_before_pixels=True)
            if p.InstanceNumber == instance_number:
                return f_path
        return None


if __name__ == '__main__':
    def main():
        selector = SliceSelector()
        selector.input_roi = os.path.join(os.environ['HOME'], 'data/surfdrive/dropbox/segmentations/vertebrae_L3.nii.gz')
        selector.input_volume = os.path.join(os.environ['HOME'], 'data/surfdrive/dropbox/nifti/al.nii.gz')
        selector.input_dicom_directory = os.path.join(os.environ['HOME'], 'data/surfdrive/dropbox/dicom')
        # selector.mode = SliceSelector.ALL
        selector.mode = 0.68
        file_paths = selector.execute()
        import shutil
        shutil.copyfile(file_paths[0], os.path.join(os.environ['HOME'], 'data/surfdrive/dropbox/L3.dcm'))
    main()



import os
import logging
import subprocess

logger = logging.getLogger(__name__)

# import dcmstack
# from dcmstack import dcmmeta


class DicomToNifti:

    def __init__(self):
        try:
            subprocess.call(['dcm2niix'])
        except FileNotFoundError:
            print(
                'dcm2niix is not installed! Please install it using the following command:\n'
                'curl -fLO https://github.com/rordenlab/dcm2niix/releases/latest/download/dcm2niix_mac.zip'
            )
        self.input_directory = None
        self.output_file = None
        self.cmd = None

    def execute(self, verbose=False):
        if self.input_directory is None:
            logger.error('Input directory not specified')
            return None
        if self.output_file is None:
            logger.error('Output file not specified')
            return None
        items = os.path.split(self.output_file)
        output_file_name = items[1]
        if output_file_name.endswith('.nii.gz'):
            output_file_name = output_file_name[:-7]
        elif output_file_name.endswith('.nii'):
            output_file_name = output_file_name[:-4]
        else:
            pass
        output_file_dir = items[0]
        os.makedirs(output_file_dir, exist_ok=True)
        self.cmd = f'dcm2niix -m y -z y -f {output_file_name} -o {output_file_dir} {self.input_directory}'
        if verbose:
            logger.info(f'{self.cmd}')
        os.system(self.cmd)
        return self.output_file


# class Dicom2Nifti:

#     def __init__(self, dcm_file_dir, nii_file_path):
#         try:
#             subprocess.call(['dcm2niix'])
#         except FileNotFoundError:
#             print(
#                 'dcm2niix is not installed! Please install it using the following command:\n'
#                 'curl -fLO https://github.com/rordenlab/dcm2niix/releases/latest/download/dcm2niix_mac.zip'
#             )
#         self.dcm_file_dir = dcm_file_dir
#         items = os.path.split(nii_file_path)
#         self.nii_file_name = items[1]
#         if self.nii_file_name.endswith('.gz'):
#             self.nii_file_name = self.nii_file_name[:-7]
#         elif self.nii_file_name.endswith('.nii'):
#             self.nii_file_name = self.nii_file_name[:-4]
#         else:
#             pass
#         self.nii_file_dir = items[0]
#         self.cmd = f'dcm2niix -m y -z y -f {self.nii_file_name} -o {self.nii_file_dir} {self.dcm_file_dir}'

#     def execute(self):
#         print(f'Executing command: {self.cmd}')
#         os.system(self.cmd)


# class Dicom2NiftiWithHeaderInfo:

#     def __init__(self, dcm_file_dir, nifti_file):
#         self.dcm_file_dir = dcm_file_dir
#         self.nifti_file = nifti_file

#     def execute(self):
#         files = [os.path.join(self.dcm_file_dir, f) for f in os.listdir(self.dcm_file_dir)]
#         # https://dcmstack.readthedocs.io/en/v0.6.1/Python_Tutorial.html#looking-up-meta-data
#         stacks = dcmstack.parse_and_stack(files, group_by='SeriesInstanceUID')
#         stack = list(stacks.values())[0]
#         nii = stack.to_nifti(embed_meta=True)
#         nii.to_filename(self.nifti_file)
#         return self.nifti_file


if __name__ == '__main__':
    def main():
        d2n = Dicom2Nifti(
            '/Users/Ralph/data/scalpel/raw/tlodewick-ct-noise-1/AL_100%/101816478/2-Abdomen',
            '/Users/Ralph/data/scalpel/processed/tlodewick-ct-noise-1-out-1/my_nifti.nii.gz',
        )
        d2n.execute()
        print(os.listdir('/Users/Ralph/data/scalpel/processed/tlodewick-ct-noise-1-out-1'))
        d2n.print_metainfo()
        # d2n = Dicom2NiftiWithHeaderInfo(
        #     '/Users/Ralph/Desktop/itk/ct-abdomen',
        #     '/Users/Ralph/Desktop/itk/out/ct-abdomen.nii.gz',
        # )
        # d2n.execute()
    main()



import os
import json
import sqlite3
import logging
import pandas as pd

logging.basicConfig()
logging.root.setLevel(logging.INFO)
logger = logging.getLogger(__name__)


class CastorQuery:

    def __init__(self, db_file, cache=True):
        self.db = self.load_db(db_file)
        self.cache = cache
        if self.cache:
            self.queries = self.read_queries_from_cache()
        else:
            self.queries = []
        self.current_query = None
        self.output = None

    def __del__(self):
        if self.db:
            self.db.close()
            self.db = None
        if self.cache:
            self.write_queries_to_cache(self.queries)

    def load_db(self, db_file):
        try:
            db = sqlite3.connect(db_file)
            return db
        except sqlite3.Error as e:
            logger.error(e)
        return None

    def read_queries_from_cache(self):
        if os.path.isfile('queries.json'):
            logger.info('Reading queries from cache...')
            with open('queries.json', 'r') as f:
                return json.load(f)
        return []

    def write_queries_to_cache(self, queries):
        logger.info('Writing queries to cache...')
        with open('queries.json', 'w') as f:
            json.dump(queries, f)

    def add_query(self, query):
        if query not in self.queries:
            self.queries.append(query)
            self.set_current_query(len(self.queries)-1)
        else:
            self.set_current_query(self.queries.index(query))

    def remove_query(self, idx):        
        self.queries.remove(self.queries[idx])
        if len(self.queries) == 0:
            self.set_current_query(None)
        else:
            self.set_current_query(0)

    def remove_all_queries(self):
        self.queries = []
        if self.cache:
            self.write_queries_to_cache(self.queries)

    def set_current_query(self, idx):
        self.current_query = self.queries[idx]

    @staticmethod
    def get_column_names(data):
        column_names = []
        for column in data.description:
            column_names.append(column[0])
        return column_names

    def execute(self):
        if self.current_query is not None and self.db is not None:
            self.output = None
            cursor = self.db.cursor()
            data = cursor.execute(self.current_query)
            df_data = []
            for result in data:
                df_data.append(result)
            self.output = pd.DataFrame(df_data, columns=self.get_column_names(data))
        return self.output

    def to_csv(self, df):
        self.output.to_csv(df, index=False, sep=';')

    def to_excel(self, df):
        self.output.to_excel(df, index=False)
                    

if __name__ == '__main__':
    def main():
        selector = CastorQuery('/Users/Ralph/Desktop/castor.db')
        # selector.set_current_query(0)
        selector.add_query('SELECT * FROM data WHERE dpca_datok BETWEEN "2018-05-01" AND "2018-07-01";')
        selector.execute()
        # selector.to_csv('query_results.csv')
        selector.to_excel('query_results.xlsx')
    main()
